/*
  nantarad-protocol. when requested, echoes back the value of 
  contents of the file (in ASCII format).
  25feb2001 Junichi Uekawa
  2 mar2001 Junichi Uekawa reused for nantarad.

  copyright 2001 Junichi Uekawa
  see GPL version 2 or later for license

  aiming to use it as a ping-echoback server to say if the machine has
  an open ppp connection.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <time.h>

#define BACKLOG 5		/* max backlog to listen to */
#define WAIT_SECONDS 10.0       /* how many seconds to wait rereading */

struct handle_incoming_params
{
  int t ;			/* socket */
  struct sockaddr_in *tmpisap;	/* address of caller */
} handle_incoming_params ;

#define malloc_with_threadexit(A,BYTES) if (NULL==((A)=malloc(BYTES)))\
{fprintf(stderr, "out of memory in allocating %i bytes\n", BYTES);pthread_exit(NULL);}

static int nantarad_value = -1 ;	/* initial value is ??? */
static pthread_mutex_t nantarad_value_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t please_read_file_cond = PTHREAD_COND_INITIALIZER;

static void handle_write ( struct handle_incoming_params * parameter)
{				/* handle writing to users */
  char sendbuf;
				/* signal file reader. */
  pthread_mutex_lock (&nantarad_value_mutex);
  sendbuf = nantarad_value;
  pthread_mutex_unlock (&nantarad_value_mutex);
  write (parameter->t, &nantarad_value, 1);
}

static void handle_incoming(struct handle_incoming_params * parameter)	/* on exit, goto cleanup */
{
  pthread_cond_broadcast (&please_read_file_cond);
  handle_write(parameter);
  /* cleanup routine */
  close(parameter->t);  
  free (parameter->tmpisap);  
  free (parameter);
}

static void read_file (char * filename)
{
				/* nantarad_value_mutex is already locked. */
  FILE* f = fopen (filename, "r");
  if (!f)
    {
      fprintf(stderr, "cannot open file %s for read\n", filename);
      return ;
    }
  fscanf (f, "%i", &nantarad_value);
  fclose(f);
}

/*
 * reads files, on request from client,
 * only when the info is 10 seconds or older.
 */
static void file_reader_thread (char * filename )
{
  time_t last_time, next_time ;
  last_time = time(NULL);
  pthread_mutex_lock (&nantarad_value_mutex);
  read_file(filename);
  
  /* be awaken by read process */
  while (1)
    {
      pthread_cond_wait (&please_read_file_cond, &nantarad_value_mutex);
      next_time=time(NULL);
      if (difftime (last_time, next_time) < -(WAIT_SECONDS))
	{
	  printf("Rereading conffile\n");
	  last_time = next_time;
	  read_file(filename) ;
	}
      else 
	{
	  printf("Not rereading conffile\n");	  
	}      
    }
  //  pthread_mutex_unlock(&nantarad_value_mutex);  
}

static char * gethostname_versatile(void)
{
  const int unithostnamelen = 256;
  int unit = 1;
  char * localhostname = malloc (unithostnamelen + 1) ;
  if (!localhostname) return NULL;
  
  while (gethostname (localhostname, unithostnamelen * unit) == -1)
    {
      char * newlocalhostname = realloc(localhostname, unithostnamelen * (++ unit) + 1);
      if (newlocalhostname == NULL)
	{
	  free(localhostname);
	  return NULL;
	}
      else 
	localhostname = newlocalhostname ;
    }
  return localhostname;
}



#define malloc_with_error(A,BYTES) if (NULL==((A)=malloc(BYTES)))\
{fprintf(stderr, "out of memory in allocating %i bytes\n", BYTES);exit(1);}

int main (int ac, char** av)
{
  char *localhostname ;
  struct sockaddr_in sa;
  struct handle_incoming_params * params;
    
  int incoming_socket;		/* the main waiting socket */
  struct hostent *hp;
  int i;
  pthread_t newthread_id ;
  pthread_attr_t pthread_attrib ;
  pthread_t filereader_id ;
  
  puts ("Nantara daemon version 0.0.1\n"
	"copyright 2001 Junichi Uekawa\n");
  if (ac != 3)
    {
      fprintf(stderr, "Please give me [port number] and [filename to watch]\n");
      exit(EXIT_FAILURE);
    }

				/* creating a pthread attribute */
  pthread_attr_init(&pthread_attrib);
  pthread_attr_setdetachstate(&pthread_attrib,1);

  
  if (!(localhostname = gethostname_versatile ()))
    {
      fprintf(stderr, "Cannot allocate enough memory to gethostname\n");
      exit(EXIT_FAILURE);
    }
  if ((hp = gethostbyname(localhostname)) == NULL)
    {
      fprintf(stderr, "%s: cannot get localhost info\n", av[0]);
      exit (EXIT_FAILURE);
    }
  if ((incoming_socket = socket (hp->h_addrtype, SOCK_STREAM, 0 )) == -1 )
    {
      perror  ("socket");
      exit (EXIT_FAILURE);      
    }

  sa.sin_port = htons(atoi(av[1]));	/* this should hopefully be portable */
  sa.sin_addr.s_addr= htonl(INADDR_ANY); /* should it be like this? it's 0 in linux */
  sa.sin_family = AF_INET;
  
  if (bind(incoming_socket, (struct sockaddr*)&sa, sizeof (sa) ) < 0  ) 
    {
      perror ("bind");
      exit (EXIT_FAILURE);
    }

  listen (incoming_socket, BACKLOG);
  if (pthread_create (&filereader_id, NULL, (void*)file_reader_thread, av[2]))
    {
      fprintf(stderr,"Thread could not be created");
      exit (EXIT_FAILURE);
    }
  /* the main routine getting incoming connections */
  while (1)
    {
      i = sizeof (struct sockaddr_in);
      malloc_with_error(params, sizeof(struct handle_incoming_params));
      malloc_with_error(params->tmpisap, i);
      if ((params->t = accept (incoming_socket, (struct sockaddr*)params->tmpisap, &i )) == -1)
	{
	  perror ("accept");
	  exit (EXIT_FAILURE);
	}
      
      pthread_create (&newthread_id, &pthread_attrib, (void*)handle_incoming, params);
    }
}
